from functools import lru_cache

from tinynetrc import Netrc

LOGIN = 'login'
PASSWORD = 'password'


class Netrc(Netrc):

    def __init__(self, file=None):
        if not file:
            from pathlib import Path

            path = Path.home() / ".netrc"
            path.touch(exist_ok=True)

        super().__init__(file=file)

@lru_cache
def get():
    return Netrc()



import html2text
from functools import lru_cache


@lru_cache
def get_parser():
    """
    
    Get an HTML2TEXT parser
     
    """
    parser = html2text.HTML2Text()
    parser.ignore_links = True
    return parser


def to_text(html):
    """

    Simple convertion a HTML document to a plain text.

    """
    parser = get_parser()
    text = parser.handle(html).strip()
    return text



from typing import Any
from yaml import CLoader as Loader, CDumper as Dumper
from yaml import load, dump


def to_yaml(obj: Any) -> str:
    """



    """
    yaml_str = dump(obj, allow_unicode=True, Dumper=Dumper)
    return yaml_str


def from_yaml(yaml_str: str) -> Any:
    """



    """
    obj = load(yaml_str, Loader=Loader)
    return obj



from diskcache import Cache

from fmtr.tools import logger, Path


class Dump(dict):
    """

    Subclass `dict` to distinguish between dumped sub-caches and regular dictionaries.

    """

class Disk(Cache):
    """

    Subclass of `diskcache` Cache that implements nested/structured caches

    """

    ROOT_KEY = '__root__'

    def __init__(self, path=None, is_root=True, **settings):
        """

        Read in existing cache structure from filesystem.

        """

        path = Path(path)
        if is_root:
            if not path.parent.exists():
                raise FileNotFoundError(f"Directory {path.parent=} does not exist")
            if path and not path.exists():
                logger.warning(f'Cache does not exist. Will be created. "{path=}"...')

            logger.info(f'Initializing Disk Cache at path "{path=}"...')

        super().__init__(directory=str(path / self.ROOT_KEY), **settings)

        self.path = path
        self.children = {}

        for path_dir in self.path.iterdir():
            if path_dir.stem == self.ROOT_KEY:
                continue
            if path_dir.is_dir():
                self.create(path_dir.name)

    def create(self, key):
        if key in self.children:
            raise KeyError(f'Sub-cache for key "{key}" already exists')
        if key in self:
            raise KeyError(f'Data for key "{key}" already exists: {repr(self[key])}')

        self.children[key] = self.__class__(self.path / key, is_root=False)

    def __getitem__(self, key):
        if key in self.children:
            return self.children[key]
        else:
            return super().__getitem__(key)

    def __setitem__(self, key, value):
        if value is type(self):
            self.create(key)
        else:
            super().__setitem__(key, value)

    def setdefault(self, key, default):
        try:
            return self[key]
        except KeyError:
            self[key] = default
            return self[key]

    def __repr__(self):
        return f'{self.__class__.__name__}("{self.path}")'

    def items(self):
        for key in self:
            yield key, self[key]

    def dump(self):
        data = Dump(self.items())
        for key, child in self.children.items():
            data[key] = child.dump()
        return data

    def iterkeys(self):
        yield from self.children.keys()
        yield from super().iterkeys()

    @property
    def data(self):
        return self.dump()


if __name__ == '__main__':
    path_tmp_cache = Path.cwd().parent.parent / 'data' / 'cache'
    tc = Disk(path_tmp_cache)

    tc.setdefault('c', Disk).setdefault('c1', Disk)['subkey'] = 0000.1
    # tc['c']=Disk.Create
    tc['c']['test'] = False
    tc['val'] = 123
    tc.setdefault('b', Disk)
    tc.setdefault('a', Disk)
    tc['a']['value2'] = 456
    tc['a']['value4'] = dict(mykey='myvalue')

    tc['b']['value3'] = [789, True]
    tc.dump()
    {}.items()



import sys

import importlib
import logging
import subprocess

from fmtr.tools import netrc_tools
from fmtr.tools.environment_tools import get

VARS = [
    (FMTR_LOG_LEVEL := 'FMTR_LOG_LEVEL'),
    (PACKAGE_NAME := 'PACKAGE_NAME'),
    (PYPI_HOST := 'PYPI_HOST'),
    (PYPI_USERNAME := 'PYPI_USERNAME'),
    (PYPI_PASSWORD := 'PYPI_PASSWORD'),
]


def run():
    FMTR_LOG_LEVEL = get('FMTR_LOG_LEVEL', default='INFO')
    logging.getLogger().setLevel(FMTR_LOG_LEVEL)

    vars = {key: get(key) for key in VARS}

    with netrc_tools.get() as netrc_obj:
        netrc_obj[vars[PYPI_HOST]] = {
            netrc_tools.LOGIN: vars[PYPI_USERNAME],
            netrc_tools.PASSWORD: vars[PYPI_PASSWORD]
        }

    command = [
        sys.executable,
        '-m',
        'pip',
        'install',
        vars[PACKAGE_NAME],
        '--no-input',
        '--index-url',
        f'https://{vars[PYPI_HOST]}'
    ]

    print(f'Starting {vars[PACKAGE_NAME]}...')

    subprocess.run(command, check=True)

    interface = importlib.import_module(f'{vars[PACKAGE_NAME]}.interface')
    interface.run()



from fmtr.tools.import_tools import MissingExtraMockModule
from fmtr.tools.inspection_tools import get_call_path

try:
    import semver

    semver = semver
    parse = semver.VersionInfo.parse
except ImportError as exception:
    parse = MissingExtraMockModule('version', exception)
    semver = MissingExtraMockModule('version', exception)


def read() -> str:
    """

    Read a generic version file from the calling package path.

    """

    from fmtr.tools.tools import ToolsConfig

    path = get_call_path(offset=2).parent / 'version'
    text = path.read_text(encoding=ToolsConfig.ENCODING).strip()

    text = get(text)

    return text


def get(text) -> str:
    """

    Optionally add dev build info.

    """
    import os
    from fmtr.tools import datatype_tools

    is_dev = datatype_tools.to_bool(os.getenv('FMTR_DEV', default=False))

    if is_dev:
        import datetime
        from fmtr.tools.tools import ToolsConfig

        timestamp = datetime.datetime.now(datetime.timezone.utc).strftime(ToolsConfig.DATETIME_SEMVER_BUILD_FORMAT)

        version = parse(text)
        version = version.bump_patch()
        version = version.replace(prerelease='dev', build=timestamp)
        text = str(version)

    return text



import flet as ft
from flet.core.event import Event
from flet.core.types import AppView
from flet.core.view import View

from fmtr.tools.logging_tools import logger


class Interface:
    """

    Simple interface base class.

    """
    TITLE = 'Base Interface'
    HOST = '0.0.0.0'
    PORT = 8080
    APPVIEW = AppView.WEB_BROWSER
    PATH_ASSETS = None
    ROUTE_ROOT = '/'

    def render(self, page: ft.Page):
        """

        Interface entry point.

        """

        if not page.on_route_change:
            page.on_route_change = lambda e, page=page: self.route(page, e)
            page.on_view_pop = lambda view, page=page: self.pop(page, view)

            page.go(self.ROUTE_ROOT)

    def route(self, page: ft.Page, event: Event):
        """

        Overridable router.

        """
        raise NotImplementedError

    def pop(self, page: ft.Page, view: View):
        """

        Overridable view pop.

        """
        raise NotImplementedError

    @classmethod
    def launch(cls):
        """

        Initialise self and launch.

        """
        self = cls()
        logger.info(f"Launching {self.TITLE} at http://{self.HOST}:{self.PORT}")
        ft.app(self.render, view=self.APPVIEW, host=self.HOST, port=self.PORT, assets_dir=self.PATH_ASSETS)


if __name__ == "__main__":
    Interface.launch()



import logging
import os

from fmtr.tools import environment_tools
from fmtr.tools.config import ToolsConfig

DEVELOPMENT = "development"
PRODUCTION = "production"
STREAM_DEFAULT = DEVELOPMENT
ENVIRONMENT_DEFAULT = DEVELOPMENT

LEVEL_DEFAULT = logging.DEBUG if environment_tools.IS_DEBUG else logging.INFO


def get_logger(name, version=None, host=ToolsConfig.FMTR_OBS_HOST, key=None, org=ToolsConfig.ORG_NAME,
               stream=STREAM_DEFAULT,
               environment=ENVIRONMENT_DEFAULT, level=LEVEL_DEFAULT):
    """

    Get a pre-configured logfire logger, if dependency is present, otherwise default to native logger.

    """

    try:
        import logfire
    except ImportError:
        logger = logging.getLogger(None)
        logger.setLevel(level)
        logger.warning(f'Logging dependencies not installed. Using native logger.')

        return logger

    if key is None:
        key = environment_tools.get(ToolsConfig.FMTR_OBS_API_KEY_KEY)
    url = f"https://{host}/api/{org}/v1/traces"
    headers = f"Authorization=Basic {key},stream-name={stream}"

    os.environ["OTEL_EXPORTER_OTLP_TRACES_ENDPOINT"] = url
    os.environ["OTEL_EXPORTER_OTLP_HEADERS"] = headers
    os.environ["OTEL_EXPORTER_OTLP_INSECURE"] = str(False).lower()

    if not version:
        from fmtr.tools import version_tools
        version = version_tools.read()

    logfire.configure(
        service_name=name,
        service_version=version,
        environment=environment,
        send_to_logfire=False,
        console=logfire.ConsoleOptions(colors='always' if environment_tools.IS_DEBUG else 'auto')
    )

    logging.getLogger(name).setLevel(level)

    logger = logfire
    return logger


logger = get_logger(name=ToolsConfig.LIBRARY_NAME)

if __name__ == '__main__':
    logger.info('Hello World')
    logger.warning('test warning')
    logger.debug('Hello World')



import re
import subprocess
from inspect import stack
from pathlib import Path
from tempfile import gettempdir
from typing import Union, Any

from fmtr.tools.config import ToolsConfig
from fmtr.tools.platform_tools import is_wsl

WIN_PATH_PATTERN = r'''([a-z]:(\\|$)|\\\\)'''
WIN_PATH_RX = re.compile(WIN_PATH_PATTERN, flags=re.IGNORECASE)


class WSLPathConversionError(EnvironmentError):
    """

    Error to raise if WSL path conversion fails.

    """


class Path(type(Path())):
    """

    Custom path object aware of WSL paths, with some additional read/write methods

    """

    def __new__(cls, *segments: Union[str, Path], convert_wsl: bool = True, **kwargs):
        """

        Intercept arguments to detect whether WSL conversion is required.

        """
        if convert_wsl and len(segments) == 1 and is_wsl() and cls.is_abs_win_path(*segments):
            segments = [cls.from_wsl(*segments)]

        return super().__new__(cls, *segments, **kwargs)

    @classmethod
    def is_abs_win_path(cls, path: Union[str, Path]) -> bool:
        """

        Infer if the current path is an absolute Windows path.

        """
        path = str(path)
        return bool(WIN_PATH_RX.match(path))

    @classmethod
    def from_wsl(cls, path: Union[str, Path]) -> bool:  # pragma: no cover
        """

        Call `wslpath` to convert the path to its Unix equivalent.

        """
        result = subprocess.run(['wslpath', '-u', str(path)], capture_output=True, text=True)

        if result.returncode:
            msg = f'Could not convert Windows path to Unix equivalent: "{path}"'
            raise WSLPathConversionError(msg)

        path_wsl = result.stdout.strip()
        path_wsl = cls(path_wsl, convert_wsl=False)
        return path_wsl

    @classmethod
    def package(cls) -> 'Path':
        """

        Get path to originating module (e.g. directory containing .py file).

        """
        frame_called = stack()[1]
        path = cls(frame_called.filename).absolute().parent
        return path

    @classmethod
    def module(cls) -> 'Path':
        """

        Get path to originating module (i.e. .py file).

        """
        frame_called = stack()[1]
        path = cls(frame_called.filename).absolute()
        return path

    @classmethod
    def temp(cls) -> 'Path':
        """

        Get path to temporary directory.

        """
        return cls(gettempdir())

    @classmethod
    def data(cls, name='data') -> 'Path':
        """

        Fetch canonical "data"/"artifacts" path, whether calling package is regular or namespace package.

        """
        from fmtr.tools.inspection_tools import get_call_path
        path = get_call_path()
        path = path.absolute().parent.parent

        path /= name

        if path.exists():
            return path

        path = path.parent.parent / name

        if path.exists():
            return path

        raise FileNotFoundError(f'No "{name}" directory found at "{path}"')

    def write_json(self, obj) -> int:
        """

        Write the specified object to the path as a JSON string

        """
        from fmtr.tools import json
        json_str = json.to_json(obj)
        return self.write_text(json_str, encoding=ToolsConfig.ENCODING)

    def read_json(self) -> Any:
        """

        Read JSON from the file and return as a Python object

        """
        from fmtr.tools import json
        json_str = self.read_text(encoding=ToolsConfig.ENCODING)
        obj = json.from_json(json_str)
        return obj

    def write_yaml(self, obj) -> int:
        """

        Write the specified object to the path as a JSON string

        """
        from fmtr.tools import yaml
        yaml_str = yaml.to_yaml(obj)
        return self.write_text(yaml_str, encoding=ToolsConfig.ENCODING)

    def read_yaml(self) -> Any:
        """

        Read YAML from the file and return as a Python object

        """
        from fmtr.tools import yaml
        yaml_str = self.read_text(encoding=ToolsConfig.ENCODING)
        obj = yaml.from_yaml(yaml_str)
        return obj

    def mkdirf(self):
        """

        Convenience method for creating directory with parents

        """
        return self.mkdir(parents=True, exist_ok=True)





import pandas as pd
from sentence_transformers import SentenceTransformer

from fmtr.tools.logging_tools import logger


class SemanticManager:
    """

    Base semantic similarity manager

    """

    REPO_ID = 'distiluse-base-multilingual-cased-v1'

    def __init__(self, data: pd.Series):
        logger.info(f"Loading model from {self.REPO_ID}")
        self.model = SentenceTransformer(self.REPO_ID)
        self.data = data
        logger.info(f"Vectorising {len(data)} texts using {self.model.device}...")
        self.embs = self.vectorise()
        logger.info(f"Vectorising complete.")

    def vectorise(self):
        """

        Vectorise the corpus

        """
        embs = self.model.encode(self.data.tolist())
        return embs

    def get_sims(self, string: str):
        """

        Get similarities between query string and corpus

        """
        logger.info(f'Getting similarities for search term: "{string}"...')
        embs_query = self.model.encode([string])
        sims = self.model.similarity(self.embs, embs_query).squeeze().numpy()
        return sims

    def get_matches(self, string: str, top_n: int = 20):
        """

        Get the Top N matches between query string and corpus

        """
        sims = self.get_sims(string)
        args = sims.argsort()[::-1]
        matches = self.data.iloc[args][:top_n]
        return matches
